
#ifndef _DRIFT_KERNEL_H_
#define _DRIFT_KERNEL_H_


// #include "doub_all.h"

/**
   This contains routines needed for a drift step for all particles
   corresponding to Hsun term described by Duncan et al. 1998
   A multiple timestep symplectic algorithms for integrating close encounters
   Duncan, M. J.,  Levison, H. F., \& Lee, M. H. 1998, AJ, 116, 2067

   To sum the momenta the parallel reduction routine by Mark Harris 
   CUDA 1.1 SDK is used.
   The citation is:
     Parallel Prefix Sum (scan) with CUDA
     Harris,  M., Sengupta, S, \& Owens, J. D.   2008, Chap 39
     in GPUGems3, edited by Hubert Nguyen,
     2008, Addison-Wesley,  Upper Saddle River, NJ, page 851

*/

// used by reduction kernel 
#ifdef __DEVICE_EMULATION__
#define EMUSYNC __syncthreads()
#else
#define EMUSYNC
#endif

///////////////////////////////////////////////////////////////////////
/**  parallel reduction Harris 07
    This version adds multiple elements per thread sequentially.  This reduces the overall
    cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
    (Brent's Theorem optimization)
    AQ: this routine should be called with 
       dim3 dimBlock(threads, 1, 1);
       dim3 dimGrid(numBlocks,1,1); but numBlocks = n/(2*threads)
   this routine meant to be run with numBlocks smaller than this by various factors of 2
       int smemSize=threads*sizeof(double);
        case 512:
            reduce6<512><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata,size); 
	break:
        case 256:
            reduce6<256><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata,size); 
*/
///////////////////////////////////////////////////////////////////////
/// g_idata input vector on device length n
/// g_odata output vector on device  length numBlocks
/// n length of vector
/// reduction kernel
template <unsigned int blockSize> // number of threads 
__global__ void
reduce6(double* g_idata, double* g_odata, unsigned int n)
{
    extern volatile __shared__ double sdata[];
    // perform first level of reduction,
    // reading from global memory, writing to shared memory
    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*(blockSize*2) + tid; 
    unsigned int gridSize = blockSize*2*gridDim.x;
    sdata[tid] = 0;

    /// we reduce multiple elements per thread.  The number is determined by the 
    /// number of active thread blocks (via gridSize).  More blocks will result
    /// in a larger gridSize and therefore fewer elements per thread
    while (i < n)
    {
        sdata[tid] += g_idata[i] + g_idata[i+blockSize];  
        i += gridSize;
    } 
    __syncthreads();

    // do reduction in shared mem
    if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
    if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
    if (blockSize >= 128) { if (tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }

#ifndef __DEVICE_EMULATION__
    if (tid < 32)
#endif
    {
        if (blockSize >=  64) { sdata[tid] += sdata[tid + 32]; EMUSYNC; }
        if (blockSize >=  32) { sdata[tid] += sdata[tid + 16]; EMUSYNC; }
        if (blockSize >=  16) { sdata[tid] += sdata[tid +  8]; EMUSYNC; }
        if (blockSize >=   8) { sdata[tid] += sdata[tid +  4]; EMUSYNC; }
        if (blockSize >=   4) { sdata[tid] += sdata[tid +  2]; EMUSYNC; }
        if (blockSize >=   2) { sdata[tid] += sdata[tid +  1]; EMUSYNC; }
    }

    
    // write result for this block to global mem 
    if (tid == 0) g_odata[blockIdx.x] = sdata[0];
    __syncthreads();
}

////////////////////////////////////////////////////////////////////
/// compute m*v for all bodies  coordpos is 0 for x part, 
///                       1 for ypart 2 for zpart
/// I did it this way so there are no case statements for x,y,z
/// however is inellegant because requires knowledge of indexing
/// includes central mass so either m0=0 or/and V0=0 should be set
/// so that this term does not contribute
//  coordpos is 0,1,2 for x,y,z
////////////////////////////////////////////////////////////////////
__global__ void
computemv(double* Pos, double* Vel, double* mv, int coordpos)
{
  int index = blockIdx.x*blockDim.x + threadIdx.x; 
  int fi = 4*index;
  mv[index] = Pos[fi + 3]*Vel[fi + coordpos];
  __syncthreads();
}

////////////////////////////////////////////////////////////////////
/// Do the drift step for all bodies
/// taking a drift step of dt/m0 sumP 
/// should be called with dt=deltaTime/(2*m0)
/// note SumP must be in global mem to not crash
////////////////////////////////////////////////////////////////////
__global__ void
driftstepBodies(double4* newPos, double4* newVel,
                double4* oldPos, double4* oldVel,
                double dt,  double3* SumP)
{
  int index = blockIdx.x*blockDim.x + threadIdx.x; 
  newPos[index] = oldPos[index]; 
  newVel[index] = oldVel[index];
  if (index >0 ){ // don't drift center of mass  (particle 0)
    newPos[index].x += dt*SumP[0].x; 
    newPos[index].y += dt*SumP[0].y;
    newPos[index].z += dt*SumP[0].z;
  }
  __syncthreads();
}


#endif // #ifndef _DRIFT_KERNEL_H_



