
#include<math.h>
#include<stdio.h>
#include "vector_types.h" // if you are running test code off gpu 
                          // you can comment this out and restore doub.h
// #include "doub.h"      

#define GPU // comment out if you are calling this subroutine without GPU
            // as in a CPU test code

// this file has a hermite integrator 
// this version allows you to keep the Sun fixed
// current version has a vector gravity routine

#define USEK    1  // compute transition functions only if USEK==1
#define NPSWITCH 50   // use gpu to do gravity if this many particles or more
#define PRINTENERGY 0 // printout energy at the beginning and end
#define NEAREST 0 // keep track of closest approaches

#define BIG 1e10

// these are some functions likely to be called externally
/*
void callHermite(int *indexlist, double4* hPos, double4 *hVel, int nindexes,
       double dtau, double eps2, double rc, double m0, int dmovesun, 
       double time, int numPlanets);
void callIntCor(int *indexlist, double4* hPos, double4* hVel, int nindexes,
       double dtau, double eps2, double rc);
double callHill(int *indexlist, int nindexes, double4 *h_Pos, double m0);
*/
// all needed subroutines for these routines seem to be present


// forward declare
double energy(double4 *h_Pos, double4 *h_Vel, int nparticles, 
     double soft2, double rc); // compute energy for checks

void kchangeover(double *kfun, double *dkdr, double *dk2dr,
          double ss , double rc); // is the transition function

double largesthill(double4 *h_Pos, int nparticles);
	// gives the largest hill radius for all particles

void findnn(double4 *h_Pos, int nparticles, int* h_cindex);
        // find nearest neighbor on cpu

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

// from gravher.cu meant to replace gravity routine here
#ifdef GPU
void gravitygpu(double4 *h_Pos, double4 *h_Vel,
                double3 *h_acc, double3 *h_jrk,
                int num, double eps2, double rc);
void findnngpu(double4 *h_Pos, int nparticles, int* h_cindex);
#endif


//////////////////////////////////////////////////////
// this routine returns accelerations (forces)
// and their time derivatives (jerks) on all particles 
// all particles reside in position,mass vector pos with velocities vel
//  acc0, jrk0 are returned
// number of particles is nparticles
// Inputs:
//    pos, vel   array of particle positions and velocities (includes masses as pow.w)
//    eps2       softening^2
//    rc         needed for transition function
//    nparticles number of particles  summing accelerations and jrks over all these
// Outputs:
//    acc0, jrk0   acceleration and jerks of all particles 
//////////////////////////////////////////////////////
void gravity(double4 *pos, double4 *vel, double3 *acc0, double3 *jrk0, 
         int nparticles, double eps2, double rc) 
{
  for(int index=0;index<nparticles;index++){
   double3 acc = {0.0,0.0,0.0};
   double3 jrk = {0.0,0.0,0.0};
   // double m0=pos[0].w;

   for(int i=0;i<nparticles;i++){
      double3 dr,dv;
      dr.x = pos[i].x - pos[index].x;
      dr.y = pos[i].y - pos[index].y;
      dr.z = pos[i].z - pos[index].z;
      dv.x = vel[i].x - vel[index].x;
      dv.y = vel[i].y - vel[index].y;
      dv.z = vel[i].z - vel[index].z;
      double dr2 = dr.x*dr.x + dr.y*dr.y + dr.z*dr.z + eps2;
      double ss  = sqrt(dr2);
      double rdotv = (dr.x*dv.x + dr.y*dv.y + dr.z*dv.z);
      double threerdotvdr2 = 3.0*rdotv/dr2;
      double inv3 = 1.0/ss/dr2; //r^-3
      double mfac = pos[i].w*inv3;
      double kfun=0.0; double dkdr=0.0; double dk2dr=0.0;
      if ((USEK==1)&&(index*i>0)) // prevent K if one index is Sun
            kchangeover(&kfun, &dkdr, &dk2dr, ss,  rc);
      
      double omk = 1.0-kfun + ss*dkdr;
      double mfac2 = mfac*omk;
      double mfac3 = rdotv*dk2dr*mfac;
      acc.x += mfac2*dr.x; 
      acc.y += mfac2*dr.y; 
      acc.z += mfac2*dr.z;
      jrk.x += mfac2*(dv.x - threerdotvdr2*dr.x) + mfac3*dr.x;
      jrk.y += mfac2*(dv.y - threerdotvdr2*dr.y) + mfac3*dr.y;
      jrk.z += mfac2*(dv.z - threerdotvdr2*dr.z) + mfac3*dr.z;
// note sign error eqn 2 of Makino+Aarseth92
// no trap for i==index as acc and jrk are zero in this case
//   this may give an error if softening length is zero
   }
   acc0[index] = acc;
   jrk0[index] = jrk;
  }
}


///////////////////////////////////////////////////////
// use 4th order hermite scheme to integrate via nbody 
// nparticles are integrated
// recipe taken from Makino, Aarseth 1992, PASJ 44, 141
// this routine takes particle positions and velocities
//   input as arrays of double4s
//   masses are in pos.w
// and integrates for a length of time tstep
// acc, jerk, and predicted pos,vel are allocated in the routine
// and then freed at end of the routine
// the computed acc and jerks are not kept
// timesteps are adjusted according to recipe
//  DONTMOVESUN used to determine whether first particle can move or not
// Inputs (and Outputs)   pos, vel   positions and velocities
//                        packed sequentially
//                        positions and velocities are updated in this routine
// Inputs:   
//    nparticles          number of particles
//    tstep               whole timestep 
//    eps2                softening squared
//    rc                  used for transition function
//    stime               external time at beginning of step
///////////////////////////////////////////////////////
#define ETA  0.001  // affects accuracy of timestep
#define ETAS 0.001  // affects initial timestep choice
#define MAXT 1e5  // a large time step
#define MINDT 1.0e-9 // smallest time step
void hermiteBodies(double4 *pos, double4 *vel, 
       int nparticles, double tstep, double eps2, double rc, int dmovesun, 
       double stime, int *indexlist, int numPlanets)
{
    double e0;
    if (PRINTENERGY==1)  {
      e0 = energy(pos, vel, nparticles, eps2, rc);
//      printf("starting hermite\n"); fflush(stdout);
    }

    double dt = MAXT;
    double dtnew = MAXT;

    int id;
    double dtc;
    double3 *acc0,*jrk0;
    double4 *posp,*velp;
    acc0 = new double3[nparticles];
    jrk0 = new double3[nparticles];
    posp = new double4[nparticles]; // predicted
    velp = new double4[nparticles];
    double3 *ag,*jg;
    ag = new double3[nparticles];
    jg = new double3[nparticles];
    

    int idstart=0; // normal
    if(dmovesun==1){ 
       idstart=1;
       posp[0] = pos[0]; // predicted are used for gravity
       velp[0] = vel[0]; // don't allow Sun to move but
                         // keep mass and position, note mass is used
    }

#ifdef GPU
    if (nparticles < NPSWITCH)
       gravity(pos,vel,ag, jg, nparticles,eps2,rc);
    else gravitygpu(pos,vel,ag, jg, nparticles,eps2,rc);
#else 
    gravity(pos,vel,ag, jg, nparticles,eps2,rc);
#endif 

// initialize and get initial timestep
    for(id = idstart;id<nparticles;id++){
   // initialize acc and jrks on particle given by index
       acc0[id] = ag[id];
       jrk0[id] = jg[id];
       double3 a0 = ag[id];
       double3 j0 = jg[id];

       dtc = sqrt(ETAS*(a0.x*a0.x + a0.y*a0.y + a0.z*a0.z)/
                         (j0.x*j0.x + j0.y*j0.y + j0.z*j0.z));
      // this is eta_s |acc/jrk| a choice for initial timestep
       dtc = MAX(dtc,MINDT); // don't allow too small
       dtc = MIN(dtc,tstep); // so final time is tstep 
                          // if initial guess is too big
   
      // initialize timestep 
       dt = MIN(dt,dtc); 
   }

   double time = 0.0;

              //////////////////////////////////
// the while loop for the integration
   while(time <tstep){

     double dt1 = dt; 
     double dt2 = dt1*dt1;
     double dt3 = dt2*dt1;
     double fac3 = dt3/6.0;
     double fac2 = dt2/2.0;
// predictor step
     for(id = idstart;id<nparticles;id++){
        posp[id].x = pos[id].x + fac3*jrk0[id].x + fac2*acc0[id].x
                     + dt1*vel[id].x;
        posp[id].y = pos[id].y + fac3*jrk0[id].y + fac2*acc0[id].y
                     + dt1*vel[id].y;
        posp[id].z = pos[id].z + fac3*jrk0[id].z + fac2*acc0[id].z
                     + dt1*vel[id].z;
        posp[id].w = pos[id].w;
        velp[id].x = vel[id].x + fac2*jrk0[id].x + dt1*acc0[id].x;
        velp[id].y = vel[id].y + fac2*jrk0[id].y + dt1*acc0[id].y;
        velp[id].z = vel[id].z + fac2*jrk0[id].z + dt1*acc0[id].z;
     }

  // update time from current timestep
      time += dt; 

// gravity step using predicted x,v
#ifdef GPU
     if (nparticles < NPSWITCH)
        gravity(posp,velp,ag, jg, nparticles,eps2,rc);  
     else gravitygpu(posp,velp,ag, jg, nparticles,eps2,rc);  
#else 
     gravity(posp,velp,ag, jg, nparticles,eps2,rc);  
#endif 
     dtnew = MAXT;
     for(id = idstart;id<nparticles;id++){
        double3 acc1 = ag[id];
        double3 jrk1 = jg[id];

// corrector step setup 
        double3 acc02, acc03, acc12;
        double3 diff;
        diff.x = acc0[id].x - acc1.x;
        diff.y = acc0[id].y - acc1.y;
        diff.z = acc0[id].z - acc1.z;
        acc02.x = -6.0*diff.x/dt2 - (4.0*jrk0[id].x + 2.0*jrk1.x)/dt1;
        acc02.y = -6.0*diff.y/dt2 - (4.0*jrk0[id].y + 2.0*jrk1.y)/dt1;
        acc02.z = -6.0*diff.z/dt2 - (4.0*jrk0[id].z + 2.0*jrk1.z)/dt1;
        acc03.x = 12.0*diff.x/dt3 + 6.0*(jrk0[id].x + jrk1.x)/dt2;
        acc03.y = 12.0*diff.y/dt3 + 6.0*(jrk0[id].y + jrk1.y)/dt2;
        acc03.z = 12.0*diff.z/dt3 + 6.0*(jrk0[id].z + jrk1.z)/dt2;
        acc12.x  = acc02.x + dt1*acc03.x; // actually used for dtnew
        acc12.y  = acc02.y + dt1*acc03.y;
        acc12.z  = acc02.z + dt1*acc03.z;

        fac3 = dt3/6.0;
        double fac4 = dt2*dt2/24.0;
        double fac5 = dt2*dt3/120.0;

// corrector step now doing correction and updating particle positions and vels
// cannot have changed predicted because other particles needed to 
// compute gravity from predicted
        pos[id].x = posp[id].x + fac4*acc02.x + fac5*acc03.x;
        pos[id].y = posp[id].y + fac4*acc02.y + fac5*acc03.y;
        pos[id].z = posp[id].z + fac4*acc02.z + fac5*acc03.z;
        vel[id].x = velp[id].x + fac3*acc02.x + fac4*acc03.x;
        vel[id].y = velp[id].y + fac3*acc02.y + fac4*acc03.y;
        vel[id].z = velp[id].z + fac3*acc02.z + fac4*acc03.z;

// now update acc0, jrk0
        acc0[id] = acc1; 
        jrk0[id] = jrk1;

// compute new timestep from all particles and in loop
      // first compute some squares of lengths of vectors
        double sacc12 = (acc12.x*acc12.x + acc12.y*acc12.y + acc12.z*acc12.z);
        double sacc13 = (acc03.x*acc03.x + acc03.y*acc03.y + acc03.z*acc03.z);
        double sacc1 = (acc1.x*acc1.x + acc1.y*acc1.y + acc1.z*acc1.z);
        double sjrk1 = (jrk1.x*jrk1.x + jrk1.y*jrk1.y + jrk1.z*jrk1.z);
        dtc = (sqrt(sacc1*sacc12) + sjrk1)/
               (sqrt(sjrk1*sacc13) + sacc12);
        dtc = sqrt(ETA*dtc);
        if (dtc > 2.0*dt1) dtc = 2.0*dt1; // don't allow to grow too fast
        dtc = MAX(dtc,MINDT); // don't let it get too small
        dtc = MIN(dtc,tstep-time);  // so final time is tstep
        dtnew = MIN(dtnew,dtc);    // dtnew should be min of all particles
      }


// choose new timestep 
      dt = dtnew;   // update step size
      dtnew  = MAXT; //reset this
   }

   delete [] acc0;
   delete [] jrk0;
   delete [] velp;
   delete [] posp;
   delete [] ag;
   delete [] jg;

   if (PRINTENERGY==1)  {
     double e1 = energy(pos, vel, nparticles, eps2, rc);
     printf("energy before and after hermite %.6e %.6e\n",e0,e1);
     fflush(stdout);
   }
//   printf("time-tstep=%.6e \n",time-tstep); not the problem
}


//////////////////////////////////////////////////
// return largest Hill radius of all particles at current radius
// the position vector, assumed to have Sun and all particles 
/////////////////////////////////////////////////
double largesthill(double4 *h_Pos, int nparticles) 
{
   double hmax = 0.0;
   double x0 = h_Pos[0].x; double y0 = h_Pos[0].y;  double z0 = h_Pos[0].z;
   double m0 = h_Pos[0].w;  
   for(int i=1;i<nparticles;i++){
      double x = h_Pos[i].x; double y = h_Pos[i].y; double z = h_Pos[i].z;
      double mi = h_Pos[i].w;  
      double r = sqrt(pow(x - x0,2.0) + pow(y - y0,2.0)+ pow(z - z0,2.0));
      double h  =  r*pow(mi/m0/3.0,1.0/3.0);
      if (h > hmax) hmax = h;
   }
   return hmax;
}

//////////////////////////////////////////////////
// return largest Hill radius of all particles at current radius 
// from an interaction list
// Inputs: 
//        indexlist     list of indexes defining particles for integration
//                      we add the Sun into the list!
//        nindexes      number of indexes not including Sun
//        h_Pos         particle positions include mass
//        m_0           mass of central body, assumed position h_Pos[0]
// Outputs: largest Hill radius of objects in list
// computed using radius from Sun not semi major axis
/////////////////////////////////////////////////
double callHill(int *indexlist, int nindexes, 
              double4 *h_Pos, double m0) 
{
   double hmax = 0.0;
   double x0=h_Pos[0].x; double y0=h_Pos[0].y; double z0=h_Pos[0].z; 
   for(int i=0;i < nindexes;i++){
      int id = indexlist[i];
      double x = h_Pos[id].x; double y = h_Pos[id].y; double z = h_Pos[id].z;
      double mi = h_Pos[id].w;  
      double r = sqrt(pow(x - x0,2.0) + pow(y - y0,2.0)+ pow(z - z0,2.0));
      double h  =  r*pow(mi/m0/3.0,1.0/3.0);
      if (h > hmax) hmax = h;
   }
   return hmax;
}
   

////////////////////////////////////////////////
// compute energy so hermite integrator can be tested
// includes energy of central body
// added softening
// Inputs:   nparticles includes cenral mass in counting
// Outputs:  total energy, using rc but not softening
#define NOM0 0 // does central mass have mass=0 and you need to correct?
               // set to 1 if you want m0 set here to 1
//////////////////////////////
double energy(double4 *h_Pos, double4 *h_Vel, int nparticles, 
         double soft2, double rc)
{
   double ke = 0.0;
   double pe = 0.0;
   double pe0 = 0.0;
   int i,j;
   // double m0=h_Pos[0].w;
  if (h_Pos[0].w ==0.0) printf("energy: m0 == 0 \n");
   for (i=0;i<nparticles;i++){
     double m  = h_Pos[i].w;
     double vx = h_Vel[i].x; double vy = h_Vel[i].y; double vz = h_Vel[i].z;
     if ((NOM0==1) && (i==0)) m=1.0;
     ke += 0.5*m*(vx*vx + vy*vy + vz*vz);
   }
   for (i=0;i<nparticles;i++){
     double x = h_Pos[i].x;  double y = h_Pos[i].y;  double z = h_Pos[i].z;
     double m  = h_Pos[i].w;
     for (j=0;j<nparticles;j++){
       double ma  = h_Pos[j].w;
       if ((NOM0==1) && (j==0)) ma=1.0;
       double xa = h_Pos[j].x;  double ya = h_Pos[j].y;  double za = h_Pos[j].z;
       double dr2 = (x-xa)*(x-xa)+ (y-ya)*(y-ya)+ (z-za)*(z-za);
       dr2 += soft2;// epsilon here
       double ss = sqrt(dr2); 
       double kfun=0.0; double dkdr=0.0; double dk2dr=0.0;
       if ((USEK==1)&&(i*j>0))  // prevent k if one index is Sun
            kchangeover(&kfun, &dkdr, &dk2dr, ss, rc);
       double omk = 1.0-kfun;
       if (i != j) pe -= (0.5*m*ma/ss)*omk;
       if ((i*j==0)&&(i!=j)) pe0 -= (0.5*m*ma/ss)*omk; // pe0 not used
     }
   }
   // printf("ke = %.6e pe = %.6e pe0=%.6e\n",ke,pe,pe0);
   return ke+pe;
}


// for energy  calculation, sum the momenta and then square it
double Edrift(double4 *hPos, double4 *hVel, int nparticles) 
{
  double3 pvec={0.0,0.0,0.0};
  double m0=hPos[0].w;
  if (NOM0==1) m0=1.0;
  for(int i=1;i<nparticles;i++){ // no sun
    pvec.x += hPos[i].w*hVel[i].x; // sum momenta
    pvec.y += hPos[i].w*hVel[i].y;
    pvec.z += hPos[i].w*hVel[i].z;
  } 
  double pp = pvec.x*pvec.x + pvec.y*pvec.y + pvec.z*pvec.z;
  double ed = pp/(2.0*m0);
  return ed;
}

// do a drift step using H_drift = |sum_i P_i|^2/(2 m0)
void driftstep (double4 *hPos, double4 *hVel, int nparticles, double dtau)
{
  double3 pvec={0.0,0.0,0.0};
  for(int i=1;i<nparticles;i++){ // sum momenta but not Sun
    double mi= hPos[i].w;
    pvec.x += mi*hVel[i].x;
    pvec.y += mi*hVel[i].y;
    pvec.z += mi*hVel[i].z;
  }
  double m0=hPos[0].w;
  double fac = dtau/m0;
  for(int i=1;i<nparticles;i++){ 
    hPos[i].x += pvec.x*fac; // move positions of all particles but Sun
    hPos[i].y += pvec.y*fac;
    hPos[i].z += pvec.z*fac;
  }
}

////////////////////////////////////////////////////////
// allocate a sub array on cpu for a few particles that will
// be integrated with Hermite scheme
// transfer data of those particles to the new arrays
// call hermiteBodies and integrate
// integrates one index list only
// transfer data back 
// Inputs: 
//        indexlist     list of indexes defining particles for integration
//                      we add the Sun into the list!
//        nindexes      number of indexes not including Sun
//        dtau          timestep
//        eps2          square of softening length
//        rc            used for transition function 
//        dmovesun      is 1 if don't move sun DONTMOVESUN
//        time          time at call of total integration
// Inputs and Outputs
//        hPos, hVel    particle and velocity arrays 
//                      are updated but only for particles in indexlist
//                      other particles are not touched
// if dmovesun==1
//   Sun's position is not changed during integration 
//   The Sun's position is assumed to be at 0,0,0  
//   and with mass m0 passed
//   hPos[0] is not used
///////////////////////////////////////////////
void callHermite(int *indexlist, double4* hPos, double4 *hVel, int nindexes, 
double dtau, double eps2, double rc, double m0,int dmovesun, double time, 
int numPlanets)
{
   double4 zero4 = {0.0, 0.0, 0.0, 0.0};
//   double m0=hPos[0].w; could be zero in Symplectic setting
   int nhermite = nindexes+1; // adding Sun
   double4 *hPoshermite, *hVelhermite; // arrays for hermite integration
   hPoshermite = new double4[nhermite];  // allocated here
   hVelhermite = new double4[nhermite];
   hPoshermite[0] = zero4; 
   hPoshermite[0].w = m0;      // here is mass of Sun
   hVelhermite[0] = zero4; 
   for(int i=0;i < nindexes;i++){
      int id = indexlist[i];
      hPoshermite[i+1] = hPos[id];  // transfer positions and mass
      hVelhermite[i+1] = hVel[id];  // and velocities
   }
   // integrate on cpu with hermite scheme
// double e0 = energy(hPoshermite, hVelhermite, nhermite, eps2, rc);
   hermiteBodies(hPoshermite, hVelhermite, nhermite, dtau, eps2,rc,
       dmovesun,time,indexlist, numPlanets);
// double e1 = energy(hPoshermite, hVelhermite, nhermite, eps2, rc);
//   printf("before after hermite energy: %.6e %.6e\n", e0,e1);
   
   for(int i=0;i < nindexes;i++){  // transfer positions and velocities back
      int id = indexlist[i];
      hPos[id] = hPoshermite[i+1];  // transfer positions  and mass
      hVel[id] = hVelhermite[i+1];  // and velocities
   }
   delete [] hPoshermite; // cleanup 
   delete [] hVelhermite;
}

////////////////////////////////////////////////////////
// Do interaction term on CPU with corrector
// This routine will change velocities only for
// sublist of particles undergoing encounters 
// Interaction step does 1
// Interaction step is supposed to have K
// Here we add (K-1) so when we add in Interaction step we get K
// Does one interaction list only
// Inputs: 
//        indexlist     list of indexes defining particles for integration
//                      not including Sun
//        dtau          timestep
//        eps2          square of softening length
//        rclist        used for transition function 
// Inputs and Outputs
//        hPos, hVel    particle and velocity arrays 
//                      are updated but only for particles in indexlist
//                      other particles are not touched
// sign has been checked
//  this was done by calling kick step and correcting and seeing that you
//    get back original
// 
///////////////////////////////////////////////
void callIntCor(int *indexlist, double4* hPos, double4* hVel, int nindexes, 
double dtau, double eps2, double rc)
{
   for(int i=0;i < nindexes;i++){ 
      int idi = indexlist[i];
      double3 acc = {0.0,0.0,0.0};
      for(int j=0;j < nindexes;j++){
         int idj = indexlist[j];
         double3 dr;
         dr.x = hPos[idi].x - hPos[idj].x;
         dr.y = hPos[idi].y - hPos[idj].y;
         dr.z = hPos[idi].z - hPos[idj].z;
         double dr2 = dr.x*dr.x + dr.y*dr.y + dr.z*dr.z + eps2;
         double ss  = sqrt(dr2);
         double inv3 = 1.0/ss/dr2; //r^-3
         double mfac = hPos[idj].w*inv3;
         double kfun=0.0; double dkdr=0.0; double dk2dr=0.0;
         if ((USEK==1)&&(idi*idj>0)) // prevent K if one index is Sun
               kchangeover(&kfun, &dkdr, &dk2dr, ss, rc);
         double omk = 1.0-kfun; // want opposite of this!
         double mfac2 =  mfac*omk; // see sign here! 
         acc.x += mfac2*dr.x; 
         acc.y += mfac2*dr.y; 
         acc.z += mfac2*dr.z;
      }
      hVel[idi].x += acc.x*dtau; // correct only velocities
      hVel[idi].y += acc.y*dtau;
      hVel[idi].z += acc.z*dtau;
   }
}



/////////////////////////////////////////////////////////////
// changeover function
// for hermite integrator we need its derivatives also
// I have a real problem with this function as it is hard to get
// radial derivatives and dependence on hill factors make interactions
// not just depend on vector between particles
//
// Inputs:
//     ss: distance between particles with smoothing length
//     rc :  a critical radius for changeover
// Outputs:
//     K    changover function
//     dk/dr its derivative w.r.t dr
//     dk2/dr its second derivative w.r.t dr
/////////////////////////////////
// #define PIOVER2 1.5707963267949
void kchangeover(double *kfun, double *dkdr,  double *dk2dr,
          double ss, double rc)
{
  *kfun = 0.0; 
  *dkdr = 0.0;
  *dk2dr = 0.0;
  if (ss==0.0) return;
  
  double y = 1.1*ss/rc - 0.1; 
  if (y<=0.0) return; // is at r<=0.0909090909 rc
  
  if (y>=1.0){ // is at r>=1 rc
    *kfun = 1.0; // *dkdr = 0.0; *dk2dr=0.0;
     return;
  }
// with cosines
  double fac = M_PI*1.1/rc;
  *kfun = 0.5*(1.0 - cos(y*M_PI));  
  *dkdr = 0.5*sin(y*M_PI)*fac;   
  *dk2dr = 0.5*cos(y*M_PI)*fac*fac;

}

////////////////////////////
// find nearest neighbors on cpu
// input: h_Pos list of posistions
//        nparticles number of particles
// output: h_cindex index list of nearest neighbors
void findnn(double4 *h_Pos, int nparticles, int* h_cindex)
{
   for(int i =0;i < nparticles; i++){
      double ldist = BIG;
      int idnn = -1;
      for(int j =0;j < nparticles; j++){ 
         if (i!=j){
            double3 dr;
            dr.x = h_Pos[i].x - h_Pos[j].x;
            dr.y = h_Pos[i].y - h_Pos[j].y;
            dr.z = h_Pos[i].z - h_Pos[j].z;
            double dr2 = dr.x*dr.x + dr.y*dr.y + dr.z*dr.z;
            if (dr2 < ldist){
               ldist = dr2;
               idnn = j;
            }
         }
      }
      h_cindex[i] = idnn;
   }
}

//////////////////////////////
// find if one particle enters the hill sphere of another
// if so print out into some file the time, distance and pairid
// now sure what I want here
//
void enterhill(double4 *pos, int nparticles, double stime)
{
   double x0 = pos[0].x;
   double y0 = pos[0].y;
   double z0 = pos[0].z;
   double m0 = pos[0].w;
   for(int i =1;i < nparticles; i++){
      double mi = pos[i].w;
      double3 dr0;
      dr0.x = pos[i].x - x0; dr0.y = pos[i].y - y0; dr0.z = pos[i].z - z0;
      double r = sqrt(dr0.x*dr0.x + dr0.y*dr0.y + dr0.z*dr0.z);
      double h = r*pow(mi/m0/3.0,1.0/3.0); // Hill radius
      for(int j =1;j < nparticles; j++){
         if (i!= j){
            double3 dr;
            dr.x = pos[i].x - pos[j].x;
            dr.y = pos[i].y - pos[j].y;
            dr.z = pos[i].z - pos[j].z;
            double dd = sqrt(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
            if (dd<h){ // particle within Hill radius
               dd*=2.0;  // what should I do? 
            }
         }
      }
   }
}
