#include <stdlib.h>
#include <stdio.h>

#include "nbody.h"

// returns lists of indexes for hermite integration
// cpu code 

// forward declare
int isinlist(int *indexlist, int nindexlist, int idi);

// call hermlists returning a vector with the first
// index of each list and the number of particles in each list
void printpairs(uint *pairlist, int npairs, int numBodies);
void printenclist(int* indexlist, int np);
void printallenclist(int* indexlist, int* startlist, int nlist);
void printtocoll(char* pstring);

//////////////////////
// on CPU sort a pairlist returning a sublists of particles 
// in encounters.
// return the number of interaction lists
// fill indexlist sequentially with particle numbers in each list
// return a vector that has the starting address of each list
// Note this routine is destructive on pairlist as it calls hermlists
// Inputs:
//    pairlist      list of pairids
//    npairs        number of pairs in pairlist
//    numBodies     needed to turn pairids into idi,idj (particle indexes)
// Outputs:
//    startlist     vector with starting positions of each list
//    indexlist     containing all lists of particles in an interaction list
//    pairlist      pairids all set to zero destructively
//    nencounter    total number of particles involved in encounters
// Returned:
//    number of interaction lists
int nhermlists(int* indexlist, int* startlist, int* nencounter,
              int numBodies, long long int* pairlist, int npairs)
{
//   printpairs(pairlist, npairs,numBodies); // if you need to print pairs
   int istart=0;
   int np=1;
   int nz = 0;
   startlist[nz] = istart;
   // int idi,idj;
   if (npairs>0){
     while(np!=0){
        np=hermlists(indexlist+istart, numBodies, pairlist, npairs);
        if (np>0){
           // printenclist(indexlist+istart, np); // if you want to print 
           nz ++;
           istart+=np;
           startlist[nz] = istart;
        }
     }
   }
   *nencounter = startlist[nz]; // total number in encounters
   return nz;

}

////////////////////////////////////////////////////
// on CPU sort a pairlist returning a single sublist of particles 
// in an encounter
// get the first list of linked particles interacting
// start with the first particle and find all interactors with it
// then find all interactors with all ones done
// You would call this until it returns zero, then there are no more pairs
// Note this routine is destructive on pairlist as each pair is removed as it 
// is placed in an encounter list
// Inputs:
//    pairlist      list of pairids
//    npairs        number of pairs in pairlist
//    numBodies     needed to turn pairids into idi,idj
// Outputs:
//    indexlist     list of particles in an interaction list
//    pairlist      pairids set to zero if their particles are put 
//                  in interaction list
// Returned 
//    number of particles in the interaction list
// Note indexlist must be large enough to contain the list of particles
//  that means at least numBodies big
//////////////////////
int hermlists(int* indexlist, int numBodies, long long int* pairlist, int npairs)
{
   int nindexlist = 0;
   int i=0;
   long long int idi,idj;
   if (npairs==0) return 0;
   while(pairlist[i]==0){ /// this will increase forever unless we check
     if (i<npairs-1) i++;
     else return 0;
   }
   if (pairlist[i]==0) return 0;
//   printf("i=%d\n",i);
//   i--;
   idi = pairlist[i]/numBodies; // this one to start it out
   idj = pairlist[i] - idi*numBodies;
   if (idi != idj){ // should never happen that idi==idj
     if (idi*idj != 0){ //get rid of sun grazers
       indexlist[nindexlist] = idi; nindexlist++; 
       indexlist[nindexlist] = idj; nindexlist++; // we now have 2 in the indexlist
     }
   }
   else printf("hermlists: idi == idj error idi=%lld \n",idi);
//   printf("idi=%d idj=%d\n",idi,idj);
   pairlist[i]=0; // zero out used ones
   int pairsfound=1;
   while(pairsfound==1){ /// keep doing this until you find all pairs
     pairsfound=0;
     for(int j=0;j<npairs;j++){
        if (pairlist[j]>0){
           idi = pairlist[j]/numBodies; 
           idj = pairlist[j] - idi*numBodies;
           int ti = isinlist(indexlist, nindexlist, idi);
           int tj = isinlist(indexlist, nindexlist, idj);
           if ((idi != idj)&&(idi*idj!=0)){ // should never happen that idi==idj
//         printf("idi=%ld ti=%d idj=%d tj=%d \n",idi,ti, idj,tj);
             if ((ti==1) && (tj==0)){ // want one in list and other not
                 pairlist[j]=0;
                 indexlist[nindexlist] = idj;
                 nindexlist++;
                 pairsfound=1;
             }
             if ((ti==0) && (tj==1)){ // one in list other not
                 pairlist[j]=0;
                 indexlist[nindexlist] = idi;
                 nindexlist++;
                 pairsfound=1;
             }
             if ((ti==1) && (tj==1)){ // both already in list
                 pairlist[j]=0; // get rid of pair
             }
           }
           else printf("hermlists: idi == idj error idi=%lld \n",idi);
        }
     } 
   }
   return nindexlist; // return the number of particles in the encounter list
  
}

// find out if particle idi already in the list 
// inputs: indexlist : the list
//         nindexlist: number of particles in list 
//         idi       : the particle idi
// outputs: return 1 if particle is in the list, 0 otherwise
int isinlist(int *indexlist, int nindexlist, int idi)
{
   for(int k=0;k<nindexlist;k++){
      if (indexlist[k] == idi) return 1; // yes, it is in list
   }
   return 0; // no, it is not in list
}

// print out some pairs
void printpairs(long long int *pairlist, int npairs, int numBodies)
{
   for(int j = 0;j<npairs;j++){
      long long int idi = pairlist[j]/numBodies; 
      long long int idj = pairlist[j] - idi*numBodies;
      printf("(%lld,%lld) ",idi,idj);
   } 
   if (npairs > 0) printf("\n");
}

// append an encounter list into coll.txt
void printenclist(int* indexlist, int np)
{
   FILE *fpa = fopen("coll.txt","a");
   int istart=0;
   fprintf(fpa,"np=%d: (",np);
   for(int i=0;i<np;i++){ 
      if (i<np-1) fprintf(fpa,"%d, ",indexlist[i+istart]);
      else  fprintf(fpa,"%d",indexlist[i+istart]);
   }
   fprintf(fpa,")\n");
   fclose(fpa);
}

// append print all encounter lists calling printenclist
void printallenclist(int* indexlist, int* startlist, int nlist)
{
   for(int i=0;i <nlist;i++){
      int offset = startlist[i];
      int nenc = startlist[i+1]-startlist[i];
      printenclist(indexlist + offset, nenc);
   }
}
// print a string to coll.txt
void printtocoll(char* pstring)
{
   FILE *fpa = fopen("coll.txt","a");
   fprintf(fpa,"%s\n",pstring); 
   fclose(fpa);
}
